@@ -6,14 +6,14 @@ Use mz tol (default 5 pmm) in XIC construction.
 XICs without neighbors within x ppm are considered specific (i.e. high selectivity). 
 Low selectivity regions will be still inspected to determine the true number of XICs.
 '''
+from operator import itemgetter
 import numpy as np
-from scipy.signal import find_peaks 
+
 from scipy import interpolate
 from scipy.ndimage import uniform_filter1d
-
 from statsmodels.nonparametric.smoothers_lowess import lowess
 
-from .mass_functions import check_close_mzs
+from .mass_functions import check_close_mzs, nn_cluster_by_mz_seeds
 
 # -----------------------------------------------------------------------------
 # mass Tracks
@@ -77,6 +77,7 @@ def extract_massTracks_(ms_expt, mz_tolerance_ppm=5, min_intensity=100, min_time
         to_remove += [a, b]
 
     updated_tracks = [tracks[ii] for ii in range(len(tracks)) if ii not in to_remove] + merged
+
     return {
         'rt_numbers': rt_numbers,
         'rt_times': rt_times,
@@ -86,12 +87,14 @@ def extract_massTracks_(ms_expt, mz_tolerance_ppm=5, min_intensity=100, min_time
 
 def extract_single_track_fullrt_length(bin, rt_length):
     '''
-    New format after v1.5.
     A mass track is an EIC for full RT range, without separating the mass traces. 
     input bins in format of [(mz_int, scan_num, intensity_int), ...].
     return a massTrack as ( mz, np.array(intensities at full rt range) ).
+    New format after v1.5.
     '''
-    mz = np.mean([x[0] for x in bin])
+    mzs = [x[0] for x in bin]
+    ints = [x[2] for x in bin]
+    mz = 0.5 * (np.median(mzs) + mzs[np.argmax(ints)])
     intensity_track = np.zeros(rt_length, dtype=np.int64)
     for r in bin:                       # this gets max intensity on the same RT scan
         intensity_track[r[1]] = max(r[2], intensity_track[r[1]])
@@ -102,37 +105,65 @@ def bin_to_mass_tracks(bin_data_tuples, rt_length, mz_tolerance_ppm=5):
     '''
     input a flexible bin by units of 0.001 amu, in format of 
                                     [(mz, scan_num, intensity_int), ...].
-    A mass track is an EIC for full RT range, without separating the mass traces. 
-    An optimization step is carried out in CMAP.MassGrid, to verify m/z separation.
+    A mass track is an EIC for full RT range, possibly containing multiple extracted ion chromatograms in traditional definition. 
 
     return massTracks as ( mz, np.array(intensities at full rt range) )
     '''
     bin_data_tuples.sort()   # by m/z, ascending
     mz_range = bin_data_tuples[-1][0] - bin_data_tuples[0][0]
     mz_tolerance = bin_data_tuples[0][0] * 0.000001 * mz_tolerance_ppm   
-    if mz_range < mz_tolerance:
+    # important: double tol_ range here as mean_mz falls in single tol_
+    if mz_range < mz_tolerance * 2:
         return [extract_single_track_fullrt_length(bin_data_tuples, rt_length)]
     else:
-        num_steps = int( 5* mz_range/ mz_tolerance )     # step in 1/5 mz_tolerance
-        hist, bin_edges = np.histogram([x[0] for x in bin_data_tuples], num_steps)
-        # example hist: array([  8,   33,  11,  24,  31,  50,  81, 164, 269,  28,   7])
-        hist_starts = [0] + [hist[:ii].sum() for ii in range(1,num_steps+1)]
-        # find_peaks returns e.g. array([ 1, 8]), {}. distance=5 because it's edge of tolerance
-        peaks, _ = find_peaks( hist, distance = 5 )
-        if peaks.any():
-            tracks = []
-            for p in peaks:
-                left = max(0, p-2)
-                right = min(p+3, num_steps)
-                tracks.append( extract_single_track_fullrt_length(
-                            bin_data_tuples[hist_starts[left]: hist_starts[right]], rt_length) )
-            return tracks
-        # possible improvement to make
-        else:
-            peak = np.argmax(hist)  # if find_peaks fails, get position of max value
-            left = max(0, peak-2)
-            right = min(peak+3, num_steps)
-            return [extract_single_track_fullrt_length(bin_data_tuples[hist_starts[left]: hist_starts[right]], rt_length)]
+        # ROIs = build_chromatogram_intensity_aware(bin_data_tuples, rt_length, mz_tolerance)
+        ROIs = build_chromatogram_by_mz_clustering(bin_data_tuples, rt_length, mz_tolerance)
+        
+        # imperfect ROIs will be examined in extract_massTracks_ and merged if needed
+        return [extract_single_track_fullrt_length(R, rt_length) for R in ROIs]
+
+
+def build_chromatogram_intensity_aware(bin_data_tuples, rt_length, mz_tolerance):
+    '''
+    Multiple m/z tracks in the same region are resolved to ROIs.
+    Start with highest intensity, going down by intensity and include data points within mz_tolerance.
+    Repeat until no track is detected. Without requiring continuous RT, which is handled in extract_single_track_fullrt_length.
+
+    Input
+    =====
+    bin_data_tuples: a flexible bin in format of [(mz, scan_num, intensity_int), ...].
+    mz_tolerance_ppm: m/z tolerance in part-per-million. Used to seggregate m/z regsions here.
+
+    return 
+    ======
+    assigned: separated bins of [(mz, scan_num, intensity_int), ...], prototype of extracted ion chromatograms 
+        to be used by extract_single_track_fullrt_length.
+    '''
+    bin_data_tuples.sort(key=itemgetter(2), reverse=True)
+    assigned, remaining = [], bin_data_tuples
+    while remaining:
+        seed, tmp_remaining = [remaining[0]], []
+        for T in remaining[1:]:
+            if abs(T[0] - seed[0][0]) < mz_tolerance:
+                seed.append(T)
+            else:
+                tmp_remaining.append(T)
+        assigned.append( seed )
+        remaining = tmp_remaining
+
+    return assigned
+
+
+def build_chromatogram_by_mz_clustering(bin_data_tuples, rt_length, mz_tolerance):
+    '''
+    Return clusters as separated bins of [(mz, scan_num, intensity_int), ...], prototype of extracted ion chromatograms 
+        to be used by extract_single_track_fullrt_length.
+    Input
+    =====
+    bin_data_tuples: a flexible bin in format of [(mz, scan_num, intensity_int), ...].
+    mz_tolerance_ppm: m/z tolerance in part-per-million. Used to seggregate m/z regsions here.
+    '''
+    return nn_cluster_by_mz_seeds(bin_data_tuples, mz_tolerance, presorted=False)
 
 
 def merge_two_mass_tracks(T1, T2):
@@ -235,7 +266,7 @@ def rt_lowess_calibration(good_landmark_peaks, selected_reference_landmark_peaks
     reference_rt_bound = max(reference_rt_numbers)
     sample_rt_bound = max(sample_rt_numbers)
     rt_rightend_ = 1.1 * sample_rt_bound
-    xx, yy = [0, 0, 0, ], [0, 0, 0, ]
+    xx, yy = [-0.1 * sample_rt_bound,]*3, [-0.1 * sample_rt_bound,]*3
     rt_cal = sorted([(x[0]['apex'], x[1]['apex']) for x in zip(good_landmark_peaks, selected_reference_landmark_peaks)])
 
     xx += [L[0] for L in rt_cal] + [rt_rightend_]*3
@@ -296,6 +327,7 @@ def remap_intensity_track(intensity_track, new, rt_cal_dict):
     Remap intensity_track based on rt_cal_dict. 
     new = basetrack.copy(), possible longer than intensity_track
     '''
+    new[ :intensity_track.shape[0]] = intensity_track.copy()
     for k,v in rt_cal_dict.items():
         new[v] = intensity_track[k]
     return new
