@@ -47,6 +47,8 @@ class ext_Experiment:
         self.parameters = parameters
         self.output_dir = parameters['outdir']
         self.mode = parameters['mode']
+        self.mz_tolerance_ppm = self.parameters['mz_tolerance_ppm']
+        self.check_isotope_ratio = self.parameters['check_isotope_ratio']
         self.database_mode = parameters['database_mode']
         self.reference_sample_id = self.get_reference_sample_id()
         
@@ -76,16 +78,21 @@ class ext_Experiment:
 
     def get_max_scan_number(self, sample_registry):
         if sample_registry:
-            return max(
-                [sample_registry[k]['max_scan_number'] for k in self.valid_sample_ids]
-            )
+            return max([sample_registry[k]['max_scan_number'] for k in self.valid_sample_ids]) + 1
         else:
             return None
 
     def process_all(self):
         '''
-        This is default asari workflow, whereas samples are aligned via LOWESS regression,
-        and peak detection is performed on composite mass tracks.
+        This is default asari workflow.
+        1. Build MassGrid, using either pairwise (small study) or clustering method. 
+            Choose one reference from all samples for the largest number of landmark m/z tracks.
+        2. RT alignment via a LOWESS function, using selective landmark peaks.
+        3. Build composite elution profile (composite_mass_tracks),
+            by cumulative sum of mass tracks from all samples after RT correction.
+        4. Global peak detection is performed on each composite massTrack.
+        5. Mapping global peaks (i.e. features) back to all samples and extract sample specific peak areas.
+            This completes the FeatureTable.
         '''
         self.CMAP = CompositeMap(self)
         self.CMAP.construct_mass_grid()
@@ -97,10 +104,8 @@ class ext_Experiment:
 
     def export_all(self):
         '''
-        initiation_Samples are used to select one most representative sample to seed MassGrid and RT alignment.
-
-        If refDB is used, it's better to be used after all samples are processed, 
-        because the m/z values of samples are closer to other samples than refDB.
+        Export all files.
+        Annotation of features to empirical compounds is done here.
         '''
         self.CMAP.MassGrid.to_csv(
             os.path.join(self.parameters['outdir'], 'export', self.parameters['mass_grid_mapping']) )
@@ -109,42 +114,57 @@ class ext_Experiment:
         self.export_feature_tables()
         self.export_log()
 
-    def annotate(self
-                                    ):
+    def annotate(self):
         '''
+        Export Feature_annotation as tsv, Annotated_empricalCompounds in both JSON and pickle.
         Reference databases can be pre-loaded.
-        Will verify ppm
+        With db_mass_calibrate to theoretical values, 
         '''
         self.load_annotation_db()
         self.db_mass_calibrate()
 
-        EED = ExperimentalEcpdDatabase(mode=self.mode)
-        EED.build_from_list_peaks(self.CMAP.FeatureList)
+        EED = ExperimentalEcpdDatabase(mode=self.mode, mz_tolerance_ppm=self.mz_tolerance_ppm)
+        EED.build_from_list_peaks(self.CMAP.FeatureList, mz_tolerance_ppm=self.mz_tolerance_ppm, check_isotope_ratio=self.check_isotope_ratio)
         EED.extend_empCpd_annotation(self.KCD)
-        EED.annotate_singletons(self.KCD)
+        EED.annotate_singletons(self.KCD)       
+        # EED.dict_empCpds misses some features 
+        EED.dict_empCpds = self.append_orphans_to_epmCpds(EED.dict_empCpds)
+
         self.export_peak_annotation(EED.dict_empCpds, self.KCD, 'Feature_annotation')
+
         if self.sample_registry:                    # check because some subcommands may not have sample_registry
-            self.select_uqnue_compound_features(EED.dict_empCpds)
+            self.select_unique_compound_features(EED.dict_empCpds)
         
-        # also exporting JSON
+        # export JSON
         outfile = os.path.join(self.parameters['outdir'], 'Annotated_empricalCompounds.json')
         with open(outfile, 'w', encoding='utf-8') as f:
             json.dump(EED.dict_empCpds, f, cls=NpEncoder, ensure_ascii=False, indent=2)
 
+        # export pickle
+        outfile = os.path.join(self.parameters['outdir'], 'export', 'epd.pickle')
+        with open(outfile, 'wb') as f:
+            pickle.dump(EED.dict_empCpds, f, pickle.HIGHEST_PROTOCOL)
+
+
     def export_CMAP_pickle(self):
         '''
-        
-        will also add MassGrid to this pickle
+        Export main CMAP data and MassGrid to pickle, which can be used for visual data exploration.
         '''
-        massTrakcs = self.CMAP.composite_mass_tracks
+        _export = {
+            '_number_of_samples_': self.CMAP._number_of_samples_,
+            'rt_length': self.CMAP.rt_length,
+            'dict_scan_rtime': self.CMAP.dict_scan_rtime,
+            'list_mass_tracks': self.CMAP.composite_mass_tracks,
+            'MassGrid': dict(self.CMAP.MassGrid),
+        }
         outfile = os.path.join(self.parameters['outdir'], 'export', 'cmap.pickle')
-
         with open(outfile, 'wb') as f:
-            pickle.dump(massTrakcs, f, pickle.HIGHEST_PROTOCOL)
+            pickle.dump(_export, f, pickle.HIGHEST_PROTOCOL)
 
 
     def load_annotation_db(self, src='hmdb4'):
-        '''Database of known compound using JMS
+        '''
+        Database of known compound using JMS. Will add more options later.
         '''
         self.KCD = knownCompoundDatabase()
         self.KCD.mass_indexed_compounds = pickle.load( pkg_resources.open_binary(db, 'mass_indexed_compounds.pickle') )
@@ -173,31 +193,34 @@ class ext_Experiment:
             print("Mass accuracy check is skipped, too few mz_landmarks (%d) matched." %len(mz_landmarks))
 
 
+    def append_orphans_to_epmCpds(self, dict_empCpds):
+        '''
+        # EED.dict_empCpds does not include features without formula match, and we add them there.
+        '''
+        all_feature_ids = []
+        for _, V in dict_empCpds.items():
+            all_feature_ids += [peak['id_number'] for peak in V['MS1_pseudo_Spectra']]
+
+        orphans = [peak for peak in self.CMAP.FeatureList if peak['id_number'] not in all_feature_ids]
+        new_id_start = len(dict_empCpds)
+        for peak in orphans:
+            dict_empCpds[new_id_start] = {'interim_id': new_id_start,
+                    'neutral_formula_mass': '', 'neutral_formula': '',
+                    'MS1_pseudo_Spectra': [peak] }
+            new_id_start += 1
+
+        return dict_empCpds
+
+
     def export_peak_annotation(self, dict_empCpds, KCD, export_file_name_prefix):
         '''
-        interim_id is empCpd id. dict_empCpds example:
-        {'interim_id': 15,
-            'neutral_formula_mass': 100.112624453,
-            'neutral_formula': 'C6H14N',
-            'Database_referred': [],
-            'identity': [],
-            'MS1_pseudo_Spectra': [{'id_number': 'F117',
-            'mz': 100.11207049661286,
-            'apex': 221.0,
-            'ion_relation': 'anchor',
-            'parent_epd_id': 15},
-            {'id_number': 'F132',
-            'mz': 101.11543204162328,
-            'apex': 221.0,
-            'ion_relation': '13C/12C',
-            'parent_epd_id': 15}],
-            'MS2_Spectra': [],
-            'list_matches': [('C6H14N_100.112624', 'M[1+]', 2),
-            ('C6H13N_99.104799', 'M+H[1+]', 2)]},...
-        '''
+        Export feature annotation.
+        interim_id is empCpd id. dict_empCpds as seen in JMS.
         
+        '''
         s = "[peak]id_number\tmz\trtime\tapex(scan number)\t[EmpCpd]interim_id\t[EmpCpd]ion_relation\tneutral_formula\tneutral_formula_mass\
         \tname_1st_guess\tmatched_DB_shorts\tmatched_DB_records\n"
+        
         for _, V in dict_empCpds.items():
             name_1st_guess, matched_DB_shorts, matched_DB_records = '', '', ''
             if 'list_matches' in V:
@@ -219,7 +242,10 @@ class ext_Experiment:
 
         print("\nAnnotation of %d Empirical compounds was written to %s.\n\n" %(len(dict_empCpds), outfile))
 
-    def select_uqnue_compound_features(self, dict_empCpds):
+    def select_unique_compound_features(self, dict_empCpds):
+        '''
+        Get unique feature by highest composite peak area per empirical compound. 
+        '''
         self.selected_unique_features = {}
         for interim_id, V in dict_empCpds.items():
             if len(V['MS1_pseudo_Spectra']) == 1:
@@ -288,10 +314,11 @@ class ext_Experiment:
                                 targeted_table.shape[0], self.number_of_samples, outfile))
 
         outfile = os.path.join(self.parameters['outdir'], 'preferred_'+self.parameters['output_feature_table'])
-        # hard coded cutoff here for now, but full table is available anyway
-        filtered_FeatureTable = filtered_FeatureTable[ filtered_FeatureTable['snr']>_snr][
-                                    filtered_FeatureTable['goodness_fitting']>_peak_shape][
-                                    filtered_FeatureTable['cSelectivity']>_cSelectivity ]
+        # Some features can have all 0s, filtered here
+        filtered_FeatureTable = filtered_FeatureTable[  filtered_FeatureTable['detection_counts'] > 0 ]
+        filtered_FeatureTable = filtered_FeatureTable[  filtered_FeatureTable['snr']>_snr][
+                                                        filtered_FeatureTable['goodness_fitting']>_peak_shape][
+                                                        filtered_FeatureTable['cSelectivity']>_cSelectivity ]
         filtered_FeatureTable.to_csv(outfile, index=False, sep="\t")
         print("Filtered Feature table (%d x %d) was written to %s.\n" %(
                                 filtered_FeatureTable.shape[0], self.number_of_samples, outfile))
@@ -312,10 +339,11 @@ class ext_Experiment:
         
     def export_log(self):
         '''
-        Export project parameters to project.json.
-        Should asari viz can look for project.json when launched standalone?
-
+        Export project parameters to project.json,
+        which is also used by asari viz.
         '''
+        self.parameters['number_of_samples'] = self.number_of_samples
+        self.parameters['number_scans'] = self.number_scans
         outfile = os.path.join(self.parameters['outdir'], 'project.json')
         with open(outfile, 'w', encoding='utf-8') as f:
             json.dump(self.parameters, f, cls=NpEncoder, ensure_ascii=False, indent=2)
