@@ -1,13 +1,16 @@
 '''
-Functions related to mass operations.
+Functions related to mass operations, inlcuding mapping and clustering.
 
-Functions here can be sped up by JIT, but the code is currently weakly typed.
+Functions here could potentially be sped up by JIT, but the code is currently weakly typed. 
 In order to use Numba for JIT, they need to be rewritten with clear typing and likely compartmentalized.
 Alternatively, some of the mass functions can be implemented in C and compiled to interface Python.
 '''
 # from numba import jit
 
 import numpy as np
+from scipy.signal import find_peaks 
+from scipy.ndimage import uniform_filter1d
+
 from mass2chem.search import *
 
 
@@ -26,24 +29,29 @@ def check_close_mzs(mzlist, ppm_tol=5):
 
     return warning
 
-# @jit(nopython=True)
 def calculate_selectivity(sorted_mz_list, std_ppm=5):
     '''
-    To calculate selectivity for all valid mass traces (thus specific m/z values).
+    To calculate m or d-selectivity for a list of m/z values, which can be all features in an experiment or a database.
 
-    The mass selectivity between two m/z values, between (0, 1), is defined as: 
+    The mass selectivity between two m/z values is defined as: 
     (1 - Probability(confusing two peaks)), further formalized as an exponential model:
-
     P = exp( -x/std_ppm ),
     whereas x is ppm distance between two peaks, 
     std_ppm standard deviation of ppm between true peaks and theoretical values, default at 5 pmm.
 
-    Close to 1 means high selectivity.
+    The selectivity value is between (0, 1), close to 1 meaning high selectivity.
     If multiple adjacent peaks are present, we multiply the selectivity scores.
-    Considering 2 lower and 2 higher neighbors approximately here.
+    It is good approximation by considering 2 lower and 2 higher neighbors here.
+    Note: ppm is actually dependent on m/z, not an ideal method. But it's in common practice and good enough.
 
-    Future direction: std_ppm can be dependent on m/z. 
-    This can be taken into account by a higher order model.
+    Input
+    =====
+    sorted_mz_list: a list of m/z values, sorted from low to high.
+    std_ppm: mass resolution in ppm (part per million).
+
+    Return
+    ======
+    A list of selectivity values, in matched order as the input m/z list.
     '''
     def __sel__(x, std_ppm=std_ppm): 
         if x > 100:         # too high, not bother
@@ -152,7 +160,7 @@ def mass_paired_mapping(list1, list2, std_ppm=5):
 
     return mapped, ratio_deltas
 
-# @jit(nopython=True)
+
 def complete_mass_paired_mapping(list1, list2, std_ppm=5):
     '''
     Similar to mass_paired_mapping, but not enforcing unique matching within std_ppm, 
@@ -304,17 +312,14 @@ def landmark_guided_mapping(REF_reference_mzlist, REF_landmarks,
     return new_reference_mzlist, new_reference_map2, REF_landmarks, _r
 
 
-
-# -----------------------------------------------------------------------------
-
-
 def seed_nn_mz_cluster(bin_data_tuples, seed_indices):
     '''
-    complete NN clustering
+    complete NN clustering, by assigning each data tuple to its closest seed.
     '''
-    seeds = [bin_data_tuples[ii] for ii in seed_indices]
-    _NN = len(seed_indices)
-    clusters = [[]] * _NN
+    seeds = [bin_data_tuples[ii] for ii in set( seed_indices) ]
+    # do set(seed_indices) to avoid repetittive seed_indices in case find_peaks does so
+    _NN = len(seeds)
+    clusters = [[]] * _NN           # see nn_cluster_by_mz_seeds for bug
     # assing cluster number by nearest distance to a seed
     for x in bin_data_tuples:
         deltas = sorted([(abs(x[0]-seeds[ii][0]), ii) for ii in range(_NN)])
@@ -324,28 +329,57 @@ def seed_nn_mz_cluster(bin_data_tuples, seed_indices):
 
 
 def gap_divide_mz_cluster(bin_data_tuples, mz_tolerance):
-    '''
-    return all clusters and each must be within mz_tolerance
-    '''
     def __divide_by_largest_gap__(L):
         gaps = [L[ii][0]-L[ii-1][0] for ii in range(1, len(L))]
         largest = np.argmax(gaps) + 1
         return L[:largest], L[largest:]
+    return __divide_by_largest_gap__(bin_data_tuples)
 
-    def __group_by_tolcheck__(good, bad, mz_tolerance):
-        tmp = []
-        for TT in bad:
-            for C in __divide_by_largest_gap__(TT):
-                # 
-                if C[-1][0] - C[0][0] < mz_tolerance:
-                    good.append(C)
-                else:
-                    tmp.append(C)
 
-        return good, tmp
+def identify_mass_peaks(bin_data_tuples, mz_tolerance, presorted=True):
+    '''
+    Get most centroid m/z values as peaks in m/z values distribution, at least mz_tolerance apart.
+    bin_data_tuples is [(mz, scan_num, intensity_int), ...] or [(m/z, track_id, sample_id), ...].
+    mz_tolerance is precomputed based on m/z and ppm, e.g. 5 ppm of 80 = 0.0004;  5 ppm of 800 = 0.0040.
+    '''
+    tol4 = int(mz_tolerance * 10000)
+    size = max(2, int(0.5 * tol4))
+    mz4 = [int(x[0]*10000) for x in bin_data_tuples]
+    if not presorted:
+        mz4.sort()
+    dict_count = {}
+    positioned = range(mz4[0], mz4[-1]+1)
+    for ii in positioned:
+        dict_count[ii] = 0
+    for x in mz4:
+        dict_count[x] += 1
+    values = uniform_filter1d([dict_count[ii] for ii in positioned], size=size, mode='nearest')
+    peaks, _ = find_peaks( values, distance = tol4 )
 
-    good, bad = [], [bin_data_tuples]
-    while bad:
-        good, bad = __group_by_tolcheck__(good, bad, mz_tolerance)
-    
-    return good
+    return [0.0001*positioned[ii] for ii in peaks]
+
+
+def nn_cluster_by_mz_seeds(bin_data_tuples, mz_tolerance, presorted=True):
+    '''
+    complete NN clustering, by assigning each data tuple to its closest m/z seed.
+    bin_data_tuples is [(mz, scan_num, intensity_int), ...] or [(m/z, track_id, sample_id), ...].
+    Note to future: np.argmin will be faster than sorted here.
+
+    Bug in Python compiler: when clusters = [[]] * _NN is used, it causes occassional duplication of entries. 2022-05-21
+
+    '''
+    mz_seeds = identify_mass_peaks(bin_data_tuples, mz_tolerance, presorted)
+    if mz_seeds:
+        _NN = len(mz_seeds)
+        # clusters = [[]] * _NN
+        clusters = []
+        for x in mz_seeds:
+            clusters.append([])
+        # assign cluster number by nearest distance to a seed
+        for x in bin_data_tuples:
+            deltas = sorted([(abs(x[0] - mz_seeds[ii]), ii) for ii in range(_NN)])
+            clusters[deltas[0][1]].append(x)
+    else:
+        clusters = [C for C in gap_divide_mz_cluster(bin_data_tuples, mz_tolerance)]
+
+    return clusters
