@@ -31,13 +31,14 @@ def iter_peak_detection_parameters(list_mass_tracks, max_rt_number, parameters,
     '''
     Prominence requirement is critical to peak detection. 
     This is dynamically determined based on parameters['min_prominence_threshold'] 
-    and intensity level in a region (min_prominence_ratio * __max_intensity).
+    and baseline level in a region.
     '''
     iters = []
     min_peak_height = parameters['min_peak_height']
     min_prominence_threshold = parameters['min_prominence_threshold']
     min_fwhm = round( 0.5 * parameters['min_timepoints'] )
     snr = parameters['signal_noise_ratio']
+    peakshape = parameters['gaussian_shape']
     wlen = 101                          # if used, this limits peak evaluation 50 scans to the left and 50 to right
     min_prominence_ratio = 0.1
     iteration = True                    # a 2nd round of peak detection if enough remaining datapoints
@@ -45,7 +46,7 @@ def iter_peak_detection_parameters(list_mass_tracks, max_rt_number, parameters,
         if mass_track['intensity'].max() > min_peak_height:
             iters.append(
                 (mass_track, max_rt_number, min_peak_height, min_fwhm, min_prominence_threshold,
-                wlen, snr, min_prominence_ratio, iteration, shared_list)
+                wlen, snr, peakshape, min_prominence_ratio, iteration, shared_list)
             )
     return iters
 
@@ -56,7 +57,7 @@ def iter_peak_detection_parameters(list_mass_tracks, max_rt_number, parameters,
 
 def stats_detect_elution_peaks(mass_track, max_rt_number, 
                 min_peak_height, min_fwhm, min_prominence_threshold,
-                wlen, snr, min_prominence_ratio, iteration,
+                wlen, snr, peakshape, min_prominence_ratio, iteration,
                 shared_list):
     '''
     Stats guided peak detection. Use statistical summary (stdev) to get ROI (region of interest), 
@@ -65,8 +66,8 @@ def stats_detect_elution_peaks(mass_track, max_rt_number,
     2) big peaks are less likely to shallow over small peaks. 
     iteration: if True, a 2nd round of peak detection if enough remaining datapoints.
 
-    No more gap allowed in peaks, since asari is using 
-    composite mass track - gaps should not exist after combining many samples.
+    Gap allowed for 1 scan. 
+    Actually in composite mass track - gaps should not exist after combining many samples.
     Now position indices have to be converted internally in this function.
 
     Input
@@ -84,53 +85,55 @@ def stats_detect_elution_peaks(mass_track, max_rt_number,
     shared_list: list of peaks in JSON format, to pool with batch_deep_detect_elution_peaks.
     '''
     list_json_peaks, list_peaks = [], []
-    list_intensity = __list_intensity = mass_track['intensity']
+    list_intensity = mass_track['intensity']
     list_scans = np.arange(max_rt_number)
+    # This baseline method down weight the peak data points
+    __baseline__ = 2**(np.log2(list_intensity + min_peak_height*0.01).mean())
+    __selected_scans__ = list_scans[list_intensity > __baseline__]
+    __noise_level__ = max(__baseline__,
+                            np.quantile( list_intensity[__selected_scans__], 0.05 ))
 
-    __number_zeros__ = len(list_scans[list_intensity < 1])
-    if list_intensity.max() > 100*min_peak_height and __number_zeros__ > 0.5*max_rt_number:
-        _do_smoothing = False           # clean mass track
-        __noise_level__ = 1
-    else:
-        _do_smoothing = True            # noisy mass track
-        __noise_level__ = min(list_intensity.std(), min_peak_height * min_prominence_ratio)     
-
-    ROIs = []                       # get ROIs by separation/filtering with __noise_level__
-    __selected_scans__ = list_scans[list_intensity > __noise_level__]
+    ROIs = []                           # get ROIs by separation/filtering with __baseline__, allowing 2 gap
     tmp = [__selected_scans__[0]]
-    for ii in __selected_scans__[1:]: # get continuous strips
-        if ii == tmp[-1] + 1:
-            tmp.append(ii)
+    for ii in __selected_scans__[1:]:
+        if ii - tmp[-1] < 3:
+            tmp += range(tmp[-1]+1, ii+1)
         else:
             ROIs.append(tmp)
             tmp = [ii]
-        
-    ROIs = [r for r in ROIs if len(r) > min_fwhm]
+
+    ROIs.append(tmp)
+    ROIs = [r for r in ROIs if len(r) >= min_fwhm * 2]
+    _do_smoothing = True                # noisy mass track
+    if list_intensity.max() > 100 * min_peak_height and len(__selected_scans__) < 0.5 * max_rt_number:
+        _do_smoothing = False           # clean mass track 
+
+    # min_prominence_ratio * list_intensity_roi.max() is not good because high noise will suppress peaks
+    prominence = max(min_prominence_threshold, snr*__baseline__)
     for R in ROIs:
+        if len(R) < 3 * min_fwhm:       # extend if too short - help narrow peaks
+            R = extend_ROI(R, max_rt_number)
         list_json_peaks += _detect_regional_peaks( list_intensity[R], R, 
-                    min_peak_height, min_fwhm, min_prominence_threshold, min_prominence_ratio, 
+                    min_peak_height, min_fwhm, prominence, min_prominence_ratio, 
                     _do_smoothing
         )
 
     # evaluate and format peaks
-    _peak_datapoints = []
-    list_cSelectivity = __peaks_cSelectivity_stats_(__list_intensity, list_json_peaks)
+    list_cSelectivity = __peaks_cSelectivity_stats_(list_intensity, list_json_peaks)
     for ii in range(len(list_json_peaks)):
         list_json_peaks[ii]['cSelectivity'] = list_cSelectivity[ii]
-        _peak_datapoints += list(range(list_json_peaks[ii]['left_base'], list_json_peaks[ii]['right_base']+1))
-
-    _noise_datapoints = [ii for ii in range(max_rt_number) if ii not in _peak_datapoints]
-    __noise_level__ = list_intensity[_noise_datapoints].mean()
+    #__noise_level__ = __baseline__
     for peak in list_json_peaks:
         peak['parent_masstrack_id'] = mass_track['id_number']
         peak['mz'] = mass_track['mz']
-        peak['snr'] = int( min(peak['height'], 99999999) / __noise_level__)         # cap upper limit and avoid INF
+        peak['snr'] = int( min(peak['height'], 99999999) / __noise_level__+1)         # cap upper limit and avoid INF
         peak['goodness_fitting'] = evaluate_gaussian_peak(mass_track, peak)
-        if peak['snr'] > snr:
+        if peak['snr'] >= snr and peak['goodness_fitting'] > peakshape:
             list_peaks.append(peak)
 
     shared_list += list_peaks
 
+
 def _detect_regional_peaks(list_intensity_roi, rt_numbers_roi, 
                     min_peak_height, min_fwhm, min_prominence_threshold, min_prominence_ratio,
                     smoothing=True):
@@ -139,20 +142,21 @@ def _detect_regional_peaks(list_intensity_roi, rt_numbers_roi,
     smooth_moving_average is applied before peak detection.
     Smoothing will reduce the height of narrow peaks in CMAP, but not on the reported values,  
     because peak area is extracted from each sample after. 
-    The likely slight expansion of peak bases adds to robustness.
+    The likely slight expansion of peak bases can add to robustness.
     '''
     list_peaks = []
     if smoothing:
-        list_intensity_roi = smooth_moving_average(list_intensity_roi, size=9)
-    prominence = max(min_prominence_threshold, min_prominence_ratio * list_intensity_roi.max())
+        list_intensity_roi = smooth_moving_average(list_intensity_roi, size=min_fwhm)
+
+    prominence = max(min_prominence_ratio * list_intensity_roi.max(), min_prominence_threshold)
     peaks, properties = find_peaks(list_intensity_roi, 
                                     height=min_peak_height, width=min_fwhm, prominence=prominence) 
-
     for ii in range(peaks.size):
         _jpeak = convert_roi_peak_json_(ii, list_intensity_roi, rt_numbers_roi, peaks, properties, )
         list_peaks.append(_jpeak)
 
     return check_overlap_peaks(list_peaks)
+    
 
 def convert_roi_peak_json_( ii, list_intensity_roi, rt_numbers_roi, peaks, properties):
     '''
@@ -178,7 +182,8 @@ def __peaks_cSelectivity_stats_(__list_intensity, _jpeaks):
     '''
     _peak_datapoints = []
     for peak in _jpeaks:
-        _peak_datapoints += list(range(peak['left_base'], peak['right_base']))
+            _peak_datapoints += list(range(peak['left_base'], peak['right_base']))
+
     _peak_datapoints = __list_intensity[list(set(_peak_datapoints))]    # peaks may overlap
     list_cSelectivity = []
     for ii in range(len(_jpeaks)):
@@ -199,7 +204,7 @@ def __peaks_cSelectivity_stats_(__list_intensity, _jpeaks):
 
 def deep_detect_elution_peaks(mass_track, max_rt_number, 
                 min_peak_height, min_fwhm, min_prominence_threshold,
-                wlen, snr, min_prominence_ratio, iteration,
+                wlen, snr, peakshape, min_prominence_ratio, iteration,
                 shared_list):
     '''
     Optimized peak detection on a mass track. 
@@ -218,7 +223,7 @@ def deep_detect_elution_peaks(mass_track, max_rt_number,
 
     Input
     =====
-    mass_track: {'id_number': k, 'mz': mz, 'rt_scan_numbers': [..], 'intensity': [..]}
+    mass_track: {'id_number': k, 'mz': mz, 'intensity': [..]}
                 Mass tracks are expected to be continuous per m/z value, with 0s for gaps.
     iteration:  Second iteration of peak detection; done on remaining data points.
                 The 2nd iteration may catch small peaks overshadowed by big ones. No clear need to go over 2.
@@ -460,7 +465,7 @@ def quick_detect_unique_elution_peak(intensity_track,
 def check_overlap_peaks(list_peaks):
     '''
     Check overlap btw a list of JSON peaks. 
-    list_peaks are already order by RT from find_peaks. Overlap usually from splitting.
+    list_peaks are already ordered by RT from find_peaks. Overlap usually from splitting.
     Return unique peaks.
     '''
     if len(list_peaks) < 2:
@@ -477,7 +482,8 @@ def check_overlap_peaks(list_peaks):
         clusters.append(tmp)
         new = []
         for C in clusters:
-            new.append(_merge_peak_cluster(C))
+            new += cleanup_peak_cluster(C)
+            
         return new
 
 def _check_overlap(peak1, peak2):
@@ -496,6 +502,11 @@ def _check_overlap(peak1, peak2):
 
 def _merge_peak_cluster(cluster_peaks):
     '''
+    Merge overlap reported peaks that stem from unclear separation.
+    If two true peaks overlap, find_peak should estabish their boundary at the valley,
+    and the reported peak bases should not overlap by N scans.
+    But good to have reevaluation of ROI if future developement desires so.
+
     Return: merged peaks by extending bases, inheritting other attributes from largest peak.
     '''
     if len(cluster_peaks) == 1:
@@ -507,3 +518,29 @@ def _merge_peak_cluster(cluster_peaks):
         largest['left_base'] = min([peak['left_base'] for peak in cluster_peaks])
         largest['right_base'] = max([peak['right_base'] for peak in cluster_peaks])
         return largest
+
+def cleanup_peak_cluster(cluster_peaks):
+    '''
+    scipy find_peaks sometimes report two overlap peak: one small and the other joined with the small peak. 
+    Separate them here. 
+    Return list of peaks.
+    '''
+    if len(cluster_peaks) == 1:
+        return cluster_peaks
+    elif len(cluster_peaks) == 2:
+        [peak1, peak2] = cluster_peaks
+        bases = list(set([peak1['left_base'], peak1['right_base'], peak2['left_base'], peak2['right_base']]))
+        bases.sort()
+        peak1['left_base'], peak1['right_base'] = bases[:2]
+        peak2['left_base'], peak2['right_base'] = bases[-2:]
+        return [peak1, peak2]
+    else:
+        return [_merge_peak_cluster(cluster_peaks)]
+
+def extend_ROI(ROI, max_rt_number):
+    '''
+    Add 3 datapoints to each end if ROI is too short (< 3*min_fwhm)
+    '''
+    left = [x for x in [ROI[0]-3, ROI[0]-2, ROI[0]-1] if x >=0]
+    right = [x for x in [ROI[-1]+1, ROI[-1]+2, ROI[-1]+3] if x < max_rt_number]
+    return left + ROI + right
